#include "main.h"

DLL_EXPORT BOOL APIENTRY DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    switch (fdwReason)
    {
    case DLL_PROCESS_ATTACH:
        /* Alloue toute les variables principales */
        /// Cration de l'event OVERLAPPED
        sAsynchIO.hEvent = CreateEvent(NULL,    // Descripteur de scurit: nul
                                       TRUE,    // RaZ manuelle
                                       FALSE,   // Non actif
                                       NULL);   // Pas de nom
        /// Erreur  la cration de l'venement
        if (sAsynchIO.hEvent == NULL)
        {
            CloseHandle(hHookFolder);
            return FALSE;
        }

        /// Allocation sur le tas (heap memory) du buffer FILE_NOTIFY_INFORMATION
        FSNotify_Size = 2 * FIELD_OFFSET(FILE_NOTIFY_INFORMATION, FileName[MAX_PATH]);
        FSNotify_Buffer = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, FSNotify_Size);

        /// Si erreur  l'allocation du buffer
        if (FSNotify_Buffer == NULL)
        {
            CloseHandle(sAsynchIO.hEvent);
            CloseHandle(hHookFolder);
            return FALSE;
        }
        break;

    case DLL_PROCESS_DETACH:
        /// Fermeture de tout les handles
        CloseMonitoring();
        break;
    }

    return TRUE;
}

int DLL_EXPORT SetMonitoringFolder(LPCTSTR HFolderPath)
{
    /// Si autre dossier dj ouvert
    if (hHookFolder != NULL)
        CloseHandle(hHookFolder);

    /// Ouvre le dossier avec tout les descripteurs requis
    hHookFolder = CreateFile(HFolderPath,
                             FILE_LIST_DIRECTORY,                                    // Parcours du dossier
                             FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, // Partage: non lock
                             NULL,                                                   // Descripteur de scurit: nul
                             OPEN_EXISTING,                                          // Doit exister
                             FILE_FLAG_BACKUP_SEMANTICS| FILE_FLAG_OVERLAPPED,       // Gre les backup (obligatoire pour ouvrir un dossier) et les transferts asynchrones
                             NULL);                                                  // Template: nul

    /// Si erreur  l'ouverture
    if (hHookFolder == INVALID_HANDLE_VALUE)
    {
        hHookFolder = NULL;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

void DLL_EXPORT SetMonitoringCallback(FSNotify_CALLBACK CBFunction)
{
    MonitoringCALLBACK = CBFunction;
}

void CloseMonitoring()
{
    HeapFree(GetProcessHeap(), 0, FSNotify_Buffer);
    CloseHandle(sAsynchIO.hEvent);
    CloseHandle(hHookFolder);
}

DWORD WINAPI MonitoringThread(LPVOID arg)
{
    while (TRUE)
    {
        GetMonitoringResult();
        Sleep((DWORD)50); // Evite l'overload du processeur
    }
}

int DLL_EXPORT GetMonitoringResult()
{
    BOOL Result;
    wchar_t CB_Param[260];

    /// Efface le prcdent buffer
    ZeroMemory(FSNotify_Buffer, FSNotify_Size);

    /// Rcupration des changements
    Result = ReadDirectoryChangesW(hHookFolder,                    // Handle du dossier
                                   FSNotify_Buffer,                // Buffer FILE_NOTIFY_INFORMATION
                                   FSNotify_Size,                  // Taille du buffer
                                   TRUE,                           // Rcursivit
                                   FILE_NOTIFY_CHANGE_FILE_NAME |  //|
                                   FILE_NOTIFY_CHANGE_DIR_NAME |   //| Fichiers ET dossier
                                   FILE_NOTIFY_CHANGE_CREATION,    //|
                                   NULL,                           // Bytes crits: non relevs
                                   &sAsynchIO,                     // Structure OVERLAPPED
                                   NULL);                          // Callback overlapped: nulle

    /// Si erreur de mise en queue
    if (!Result)
    {
        CloseMonitoring();
        return 1;
    }

    /// Rcupration des infos de transfert asynchrone
    DWORD dwRetBytes;
    Result = GetOverlappedResult(hHookFolder,  // Handle du dossier
                                 &sAsynchIO,   // Structure OVERLAPPED
                                 &dwRetBytes,  // Nombre d'octets crits dans la structure
                                 TRUE);        // Mode synchrone

    /// Si erreur de transfert asynchrone
    if (!Result)
    {
        CloseMonitoring();
        return 2;
    }

    /// Transfert du buffer dans la structure FILE_NOTIFY_INFORMATION
    sFSNotify = FSNotify_Buffer;
    dwRetBytes = min(dwRetBytes, FSNotify_Size);

    /// Tant qu'il y a des changements dans la file
    while (dwRetBytes >= FIELD_OFFSET(FILE_NOTIFY_INFORMATION, FileName[sFSNotify->FileNameLength / sizeof(WCHAR)]))
    {

        /// Cration de la description et envois  la callback
        swprintf(CB_Param, L"%d%*ls",  (int)sFSNotify->Action, sFSNotify->FileNameLength / sizeof(WCHAR), sFSNotify->FileName);
        (MonitoringCALLBACK)(CB_Param);


        /// Si pas de changement suivant dans la file
        if (sFSNotify->NextEntryOffset == 0)
            break;

        /// Si on a lu toute les octets 'utiles' disponibles
        if (dwRetBytes < sFSNotify->NextEntryOffset)
            break;

        /// MaJ du nombre d'octets restant  lire et migration de la structure depuis le buffer
        dwRetBytes -= sFSNotify->NextEntryOffset;
        sFSNotify = (PVOID)((ULONG_PTR)sFSNotify + sFSNotify->NextEntryOffset);
    }

    /// RaZ de l'venement asynchrone
    ZeroMemory(&sAsynchIO, FIELD_OFFSET(OVERLAPPED, hEvent));
    ResetEvent(sAsynchIO.hEvent);

    return EXIT_SUCCESS;
}

void DLL_EXPORT StartThreadMonitoring()
{
    /* Demarre le thread de surveillance */
    CreateThread(NULL, 0, &MonitoringThread, NULL, 0, ThreadID);
}

void DLL_EXPORT StopThreadMonitoring()
{
    TerminateThread(ThreadID, (DWORD)0);
}
